@@ -1,42 +1,6 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 Klarälvdalens Datakonsult AB, a KDAB Group company, info@kdab.com, author Milian Wolff <milian.wolff@kdab.com>
-** Copyright (C) 2019 Menlo Systems GmbH, author Arno Rehn <a.rehn@menlosystems.com>
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtWebChannel module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 Klarälvdalens Datakonsult AB, a KDAB Group company, info@kdab.com, author Milian Wolff <milian.wolff@kdab.com>
+// Copyright (C) 2019 Menlo Systems GmbH, author Arno Rehn <a.rehn@menlosystems.com>
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "qmetaobjectpublisher_p.h"
 #include "qwebchannel.h"
@@ -44,6 +8,9 @@
 #include "qwebchannelabstracttransport.h"
 
 #include <QEvent>
+#if QT_CONFIG(future)
+#include <QFuture>
+#endif
 #include <QJsonDocument>
 #include <QDebug>
 #include <QJsonObject>
@@ -78,7 +45,7 @@ bool isQFlagsType(uint id)
         return false;
     }
 
-    QByteArray name = QMetaType::typeName(id);
+    QByteArray name = type.name();
     name = name.mid(name.lastIndexOf(":") + 1);
     return mo->indexOfEnumerator(name.constData()) > -1;
 }
@@ -123,7 +90,7 @@ int doubleToNumberConversionScore(int userType)
         break;
     }
 
-    if (QMetaType::typeFlags(userType) & QMetaType::IsEnumeration)
+    if (QMetaType(userType).flags() & QMetaType::IsEnumeration)
         return doubleToNumberConversionScore(QMetaType::Int);
 
     return IncompatibleScore;
@@ -177,19 +144,112 @@ QJsonObject createResponse(const QJsonValue &id, const QJsonValue &data)
     return response;
 }
 
-/// TODO: what is the proper value here?
-const int PROPERTY_UPDATE_INTERVAL = 50;
+#if QT_CONFIG(future)
+QMetaType resultTypeOfQFuture(QByteArrayView typeName)
+{
+    if (!typeName.startsWith("QFuture<") || !typeName.endsWith('>'))
+        return {};
+
+    return QMetaType::fromName(typeName.sliced(8, typeName.size() - 9));
 }
 
-Q_DECLARE_TYPEINFO(OverloadResolutionCandidate, Q_MOVABLE_TYPE);
+template<typename Func>
+void attachContinuationToFutureInVariant(const QVariant &result, QPointer<QObject> contextObject,
+                                         Func continuation)
+{
+    Q_ASSERT(result.canConvert<QFuture<void>>());
+
+    // QMetaObject::invokeMethod() indirection to work around an issue with passing
+    // a context object to QFuture::then(). See below.
+    const auto safeContinuation = [contextObject, continuation=std::move(continuation)]
+            (const QVariant &result)
+    {
+        if (!contextObject)
+            return;
+
+        QMetaObject::invokeMethod(contextObject.get(), [continuation, result] {
+            continuation(result);
+        });
+    };
+
+    auto f = result.value<QFuture<void>>();
+
+    // Be explicit about what we capture so that we don't accidentally run into
+    // threading issues.
+    f.then([resultType=resultTypeOfQFuture(result.typeName()), f, continuation=safeContinuation]
+    {
+        if (!resultType.isValid() || resultType == QMetaType::fromType<void>()) {
+            continuation(QVariant{});
+            return;
+        }
+
+        auto iface = QFutureInterfaceBase::get(f);
+        // If we pass a context object to f.then() and the future originates in a
+        // different thread, this assertions fails. Why?
+        // For the time being, work around that with QMetaObject::invokeMethod()
+        // in safeSendResponse().
+        Q_ASSERT(iface.resultCount() > 0);
+
+        QMutexLocker<QMutex> locker(&iface.mutex());
+        if (iface.resultStoreBase().resultAt(0).isVector()) {
+            locker.unlock();
+            // This won't work because we cannot generically get a QList<T> into
+            // a QVariant with T only known at runtime.
+            // TBH, I don't know how to trigger this.
+            qWarning() << "Result lists in a QFuture return value are not supported!";
+            continuation(QVariant{});
+            return;
+        }
+
+        // pointer<void>() wouldn't compile because of the isVector-codepath
+        // using QList<T> in that method. We're not taking that path anyway (see the
+        // above check), so we can use char instead to not break strict aliasing
+        // requirements.
+        const auto data = iface.resultStoreBase().resultAt(0).pointer<char>();
+        locker.unlock();
+
+        const QVariant result(resultType, data);
+        continuation(result);
+    }).onCanceled([continuation=safeContinuation] {
+        // Will catch both failure and cancellation.
+        // Maybe send something more meaningful?
+        continuation(QVariant{});
+    });
+}
+#endif
+
+}
+
+Q_DECLARE_TYPEINFO(OverloadResolutionCandidate, Q_RELOCATABLE_TYPE);
+
+void QWebChannelPropertyChangeNotifier::notify(
+        QPropertyObserver *self, QUntypedPropertyData *)
+{
+    auto This = static_cast<QWebChannelPropertyChangeNotifier*>(self);
+
+    // Use the indirection with Qt::AutoConnection to ensure invocation
+    // in the correct thread.
+    // Explicitly copy the parameters into the lambda so that this instance can be destroyed after posting a queued
+    // invocation. The current design doesn't allow this anyway, but I don't want bad surprises in a possible future
+    // commit.
+    QMetaObject::invokeMethod(
+                This->publisher,
+                [publisher=This->publisher, object=This->object, propertyIndex=This->propertyIndex]
+    {
+        publisher->propertyValueChanged(object, propertyIndex);
+    }, Qt::AutoConnection);
+}
 
 QMetaObjectPublisher::QMetaObjectPublisher(QWebChannel *webChannel)
-    : QObject(webChannel)
-    , webChannel(webChannel)
-    , signalHandler(this)
-    , clientIsIdle(false)
-    , blockUpdates(false)
-    , propertyUpdatesInitialized(false)
+    : QObject(webChannel),
+      webChannel(webChannel),
+      blockUpdatesStatus(false),
+      blockUpdatesHandler(blockUpdatesStatus.onValueChanged(
+              std::function<void()>([&]() { this->onBlockUpdatesChanged(); }))),
+      propertyUpdatesInitialized(false),
+      propertyUpdateIntervalTime(50),
+      propertyUpdateIntervalHandler(propertyUpdateIntervalTime.onValueChanged(
+              std::function<void()>([&]() { this->startPropertyUpdateTimer(true); })))
 {
 }
 
@@ -207,7 +267,7 @@ void QMetaObjectPublisher::registerObject(const QString &id, QObject *object)
             qWarning("Registered new object after initialization, existing clients won't be notified!");
             // TODO: send a message to clients that an object was added
         }
-        initializePropertyUpdates(object, classInfoForObject(object, Q_NULLPTR));
+        initializePropertyUpdates(object, classInfoForObject(object, nullptr));
     }
 }
 
@@ -240,7 +300,7 @@ QJsonObject QMetaObjectPublisher::classInfoForObject(const QObject *object, QWeb
             // optimize: compress the common propertyChanged notification names, just send a 1
             const QByteArray &notifySignal = prop.notifySignal().name();
             static const QByteArray changedSuffix = QByteArrayLiteral("Changed");
-            if (notifySignal.length() == changedSuffix.length() + propertyName.length() &&
+            if (notifySignal.size() == changedSuffix.size() + propertyName.size() &&
                 notifySignal.endsWith(changedSuffix) && notifySignal.startsWith(prop.name()))
             {
                 signalInfo.append(1);
@@ -248,8 +308,8 @@ QJsonObject QMetaObjectPublisher::classInfoForObject(const QObject *object, QWeb
                 signalInfo.append(QString::fromLatin1(notifySignal));
             }
             signalInfo.append(prop.notifySignalIndex());
-        } else if (!prop.isConstant()) {
-            qWarning("Property '%s'' of object '%s' has no notify signal and is not constant, "
+        } else if (!prop.isConstant() && !prop.isBindable()) {
+            qWarning("Property '%s'' of object '%s' has no notify signal, is not bindable and is not constant, "
                      "value updates in HTML will be broken!",
                      prop.name(), object->metaObject()->className());
         }
@@ -301,17 +361,17 @@ QJsonObject QMetaObjectPublisher::classInfoForObject(const QObject *object, QWeb
     return data;
 }
 
-void QMetaObjectPublisher::setClientIsIdle(bool isIdle)
+void QMetaObjectPublisher::setClientIsIdle(bool isIdle, QWebChannelAbstractTransport *transport)
 {
-    if (clientIsIdle == isIdle) {
-        return;
-    }
-    clientIsIdle = isIdle;
-    if (!isIdle && timer.isActive()) {
-        timer.stop();
-    } else if (isIdle && !timer.isActive()) {
-        timer.start(PROPERTY_UPDATE_INTERVAL, this);
-    }
+    transportState[transport].clientIsIdle = isIdle;
+    if (isIdle)
+        sendEnqueuedPropertyUpdates(transport);
+}
+
+bool QMetaObjectPublisher::isClientIdle(QWebChannelAbstractTransport *transport)
+{
+    auto found = transportState.find(transport);
+    return found != transportState.end() && found.value().clientIsIdle;
 }
 
 QJsonObject QMetaObjectPublisher::initializeClient(QWebChannelAbstractTransport *transport)
@@ -331,9 +391,11 @@ QJsonObject QMetaObjectPublisher::initializeClient(QWebChannelAbstractTransport
     return objectInfos;
 }
 
-void QMetaObjectPublisher::initializePropertyUpdates(const QObject *const object, const QJsonObject &objectInfo)
+void QMetaObjectPublisher::initializePropertyUpdates(QObject *const object, const QJsonObject &objectInfo)
 {
-    foreach (const QJsonValue &propertyInfoVar, objectInfo[KEY_PROPERTIES].toArray()) {
+    auto *metaObject = object->metaObject();
+    auto *signalHandler = signalHandlerFor(object);
+    for (const auto propertyInfoVar : objectInfo[KEY_PROPERTIES].toArray()) {
         const QJsonArray &propertyInfo = propertyInfoVar.toArray();
         if (propertyInfo.size() < 2) {
             qWarning() << "Invalid property info encountered:" << propertyInfoVar;
@@ -341,31 +403,41 @@ void QMetaObjectPublisher::initializePropertyUpdates(const QObject *const object
         }
         const int propertyIndex = propertyInfo.at(0).toInt();
         const QJsonArray &signalData = propertyInfo.at(2).toArray();
+        const QMetaProperty metaProp = metaObject->property(propertyIndex);
 
-        if (signalData.isEmpty()) {
-            // Property without NOTIFY signal
-            continue;
+        if (!signalData.isEmpty()) {
+            // Property with a NOTIFY signal
+            const int signalIndex = signalData.at(1).toInt();
+
+            QSet<int> &connectedProperties = signalToPropertyMap[object][signalIndex];
+
+            // Only connect for a property update once
+            if (connectedProperties.isEmpty()) {
+                signalHandler->connectTo(object, signalIndex);
+            }
+
+            connectedProperties.insert(propertyIndex);
+        } else if (metaProp.isBindable()) {
+            const auto [begin, end] = propertyObservers.equal_range(object);
+            const auto it = std::find_if(begin, end, [&](auto &n) {
+                return n.second.propertyIndex == propertyIndex;
+            });
+            // Only connect for a property update once
+            if (it == end) {
+                auto it = propertyObservers.emplace(
+                            object, QWebChannelPropertyChangeNotifier{this, object, propertyIndex});
+                metaProp.bindable(object).observe(&it->second);
+            }
         }
-
-        const int signalIndex = signalData.at(1).toInt();
-
-        QSet<int> &connectedProperties = signalToPropertyMap[object][signalIndex];
-
-        // Only connect for a property update once
-        if (connectedProperties.isEmpty()) {
-            signalHandler.connectTo(object, signalIndex);
-        }
-
-        connectedProperties.insert(propertyIndex);
     }
 
     // also always connect to destroyed signal
-    signalHandler.connectTo(object, s_destroyedSignalIndex);
+    signalHandler->connectTo(object, s_destroyedSignalIndex);
 }
 
 void QMetaObjectPublisher::sendPendingPropertyUpdates()
 {
-    if (blockUpdates || !clientIsIdle || pendingPropertyUpdates.isEmpty()) {
+    if (blockUpdatesStatus) {
         return;
     }
 
@@ -383,16 +455,21 @@ void QMetaObjectPublisher::sendPendingPropertyUpdates()
         QJsonObject properties;
         // maps signal index to list of arguments of the last emit
         QJsonObject sigs;
-        const SignalToArgumentsMap::const_iterator sigEnd = it.value().constEnd();
-        for (SignalToArgumentsMap::const_iterator sigIt = it.value().constBegin(); sigIt != sigEnd; ++sigIt) {
-            // TODO: can we get rid of the int <-> string conversions here?
-            foreach (const int propertyIndex, objectsSignalToPropertyMap.value(sigIt.key())) {
-                const QMetaProperty &property = metaObject->property(propertyIndex);
-                Q_ASSERT(property.isValid());
-                properties[QString::number(propertyIndex)] = wrapResult(property.read(object), Q_NULLPTR, objectId);
-            }
+
+        const auto indexes = it.value().propertyIndices(objectsSignalToPropertyMap);
+
+        // TODO: can we get rid of the int <-> string conversions here?
+        for (const int propertyIndex : indexes) {
+            const QMetaProperty &property = metaObject->property(propertyIndex);
+            Q_ASSERT(property.isValid());
+            properties[QString::number(propertyIndex)] = wrapResult(property.read(object), nullptr, objectId);
+        }
+
+        const auto sigMap = it.value().signalMap;
+        for (auto sigIt = sigMap.constBegin(); sigIt != sigMap.constEnd(); ++sigIt) {
             sigs[QString::number(sigIt.key())] = QJsonArray::fromVariantList(sigIt.value());
         }
+
         QJsonObject obj;
         obj[KEY_OBJECT] = objectId;
         obj[KEY_SIGNALS] = sigs;
@@ -400,7 +477,7 @@ void QMetaObjectPublisher::sendPendingPropertyUpdates()
 
         // if the object is auto registered, just send the update only to clients which know this object
         if (wrappedObjects.contains(objectId)) {
-            foreach (QWebChannelAbstractTransport *transport, wrappedObjects.value(objectId).transports) {
+            for (QWebChannelAbstractTransport *transport : wrappedObjects.value(objectId).transports) {
                 QJsonArray &arr = specificUpdates[transport];
                 arr.push_back(obj);
             }
@@ -415,18 +492,19 @@ void QMetaObjectPublisher::sendPendingPropertyUpdates()
 
     // data does not contain specific updates
     if (!data.isEmpty()) {
-        setClientIsIdle(false);
-
         message[KEY_DATA] = data;
-        broadcastMessage(message);
+        enqueueBroadcastMessage(message);
     }
 
     // send every property update which is not supposed to be broadcasted
     const QHash<QWebChannelAbstractTransport*, QJsonArray>::const_iterator suend = specificUpdates.constEnd();
     for (QHash<QWebChannelAbstractTransport*, QJsonArray>::const_iterator it = specificUpdates.constBegin(); it != suend; ++it) {
         message[KEY_DATA] = it.value();
-        it.key()->sendMessage(message);
+        enqueueMessage(message, it.key());
     }
+
+    for (auto state = transportState.begin(); state != transportState.end(); ++state)
+        sendEnqueuedPropertyUpdates(state.key());
 }
 
 QVariant QMetaObjectPublisher::invokeMethod(QObject *const object, const QMetaMethod &method,
@@ -457,6 +535,7 @@ QVariant QMetaObjectPublisher::invokeMethod(QObject *const object, const QMetaMe
     VariantArgument arguments[10];
     for (int i = 0; i < qMin(args.size(), method.parameterCount()); ++i) {
         arguments[i].value = toVariant(args.at(i), method.parameterType(i));
+        arguments[i].type = method.parameterType(i);
     }
     // construct QGenericReturnArgument
     QVariant returnValue;
@@ -470,7 +549,7 @@ QVariant QMetaObjectPublisher::invokeMethod(QObject *const object, const QMetaMe
         // Only init variant with return type if its not a variant itself, which would
         // lead to nested variants which is not what we want.
         if (method.returnType() != QMetaType::QVariant)
-            returnValue = QVariant(method.returnType(), 0);
+            returnValue = QVariant(QMetaType(method.returnType()), nullptr);
 
         QGenericReturnArgument returnArgument(method.typeName(), returnValue.data());
         method.invoke(object, returnArgument,
@@ -496,7 +575,7 @@ QVariant QMetaObjectPublisher::invokeMethod(QObject *const object, const int met
 QVariant QMetaObjectPublisher::invokeMethod(QObject *const object, const QByteArray &methodName,
                                             const QJsonArray &args)
 {
-    QVector<OverloadResolutionCandidate> candidates;
+    QList<OverloadResolutionCandidate> candidates;
 
     const QMetaObject *mo = object->metaObject();
     for (int i = 0; i < mo->methodCount(); ++i) {
@@ -554,13 +633,13 @@ void QMetaObjectPublisher::signalEmitted(const QObject *object, const int signal
         message[KEY_OBJECT] = objectName;
         message[KEY_SIGNAL] = signalIndex;
         if (!arguments.isEmpty()) {
-            message[KEY_ARGS] = wrapList(arguments, Q_NULLPTR, objectName);
+            message[KEY_ARGS] = wrapList(arguments, nullptr, objectName);
         }
         message[KEY_TYPE] = TypeSignal;
 
         // if the object is wrapped, just send the response to clients which know this object
         if (wrappedObjects.contains(objectName)) {
-            foreach (QWebChannelAbstractTransport *transport, wrappedObjects.value(objectName).transports) {
+            for (QWebChannelAbstractTransport *transport : wrappedObjects.value(objectName).transports) {
                 transport->sendMessage(message);
             }
         } else {
@@ -571,10 +650,28 @@ void QMetaObjectPublisher::signalEmitted(const QObject *object, const int signal
             objectDestroyed(object);
         }
     } else {
-        pendingPropertyUpdates[object][signalIndex] = arguments;
-        if (clientIsIdle && !blockUpdates && !timer.isActive()) {
-            timer.start(PROPERTY_UPDATE_INTERVAL, this);
-        }
+        auto &propertyUpdate = pendingPropertyUpdates[object];
+        propertyUpdate.signalMap[signalIndex] = arguments;
+        startPropertyUpdateTimer();
+    }
+}
+
+void QMetaObjectPublisher::propertyValueChanged(const QObject *object, const int propertyIndex)
+{
+    pendingPropertyUpdates[object].plainProperties.insert(propertyIndex);
+    startPropertyUpdateTimer();
+}
+
+void QMetaObjectPublisher::startPropertyUpdateTimer(bool forceRestart)
+{
+    if (blockUpdatesStatus)
+        return;
+
+    if (propertyUpdateIntervalTime >= 0) {
+        if (forceRestart || !timer.isActive())
+            timer.start(propertyUpdateIntervalTime, this);
+    } else {
+        sendPendingPropertyUpdates();
     }
 }
 
@@ -590,10 +687,11 @@ void QMetaObjectPublisher::objectDestroyed(const QObject *object)
     // only remove from handler when we initialized the property updates
     // cf: https://bugreports.qt.io/browse/QTBUG-60250
     if (propertyUpdatesInitialized) {
-        signalHandler.remove(object);
+        signalHandlerFor(object)->remove(object);
         signalToPropertyMap.remove(object);
     }
     pendingPropertyUpdates.remove(object);
+    propertyObservers.erase(object);
 }
 
 QObject *QMetaObjectPublisher::unwrapObject(const QString &objectId) const
@@ -608,11 +706,50 @@ QObject *QMetaObjectPublisher::unwrapObject(const QString &objectId) const
     }
 
     qWarning() << "No wrapped object" << objectId;
-    return Q_NULLPTR;
+    return nullptr;
+}
+
+QVariant QMetaObjectPublisher::unwrapMap(QVariantMap map) const
+{
+    const auto qobj = map.value(KEY_QOBJECT).toBool();
+    const auto id = qobj ? map.value(KEY_ID).toString() : QString();
+
+    if (!id.isEmpty()) // it's probably a QObject
+        return QVariant::fromValue(unwrapObject(id));
+
+    // it's probably just a normal JS object, continue searching for objects
+    // that look like QObject*
+    for (auto &value : map)
+        value = unwrapVariant(value);
+
+    return map;
+}
+
+QVariant QMetaObjectPublisher::unwrapList(QVariantList list) const
+{
+    for (auto &value : list)
+        value = unwrapVariant(value);
+
+    return list;
+}
+
+QVariant QMetaObjectPublisher::unwrapVariant(const QVariant &value) const
+{
+    switch (value.metaType().id())
+    {
+        case QMetaType::QVariantList:
+            return unwrapList(value.toList());
+        case QMetaType::QVariantMap:
+            return unwrapMap(value.toMap());
+        default:
+            break;
+    }
+    return value;
 }
 
 QVariant QMetaObjectPublisher::toVariant(const QJsonValue &value, int targetType) const
 {
+    QMetaType target(targetType);
     if (targetType == QMetaType::QJsonValue) {
         return QVariant::fromValue(value);
     } else if (targetType == QMetaType::QJsonArray) {
@@ -623,34 +760,35 @@ QVariant QMetaObjectPublisher::toVariant(const QJsonValue &value, int targetType
         if (!value.isObject())
             qWarning() << "Cannot not convert non-object argument" << value << "to QJsonObject.";
         return QVariant::fromValue(value.toObject());
-    } else if (QMetaType::typeFlags(targetType) & QMetaType::PointerToQObject) {
+    } else if (target.flags() & QMetaType::PointerToQObject) {
         QObject *unwrappedObject = unwrapObject(value.toObject()[KEY_ID].toString());
-        if (unwrappedObject == Q_NULLPTR)
+        if (unwrappedObject == nullptr)
             qWarning() << "Cannot not convert non-object argument" << value << "to QObject*.";
         return QVariant::fromValue(unwrappedObject);
     } else if (isQFlagsType(targetType)) {
         int flagsValue = value.toInt();
-        return QVariant(targetType, reinterpret_cast<const void*>(&flagsValue));
+        return QVariant(target, reinterpret_cast<const void*>(&flagsValue));
     }
 
     // this converts QJsonObjects to QVariantMaps, which is not desired when
     // we want to get a QJsonObject or QJsonValue (see above)
-    QVariant variant = value.toVariant();
-    if (targetType != QMetaType::QVariant && !variant.convert(targetType)) {
-        qWarning() << "Could not convert argument" << value << "to target type" << QVariant::typeToName(targetType) << '.';
+    QVariant variant = unwrapVariant(value.toVariant());
+    if (targetType != QMetaType::QVariant && !variant.convert(target)) {
+        qWarning() << "Could not convert argument" << value << "to target type" << target.name() << '.';
     }
     return variant;
 }
 
 int QMetaObjectPublisher::conversionScore(const QJsonValue &value, int targetType) const
 {
+    QMetaType target(targetType);
     if (targetType == QMetaType::QJsonValue) {
         return PerfectMatchScore;
     } else if (targetType == QMetaType::QJsonArray) {
         return value.isArray() ? PerfectMatchScore : IncompatibleScore;
     } else if (targetType == QMetaType::QJsonObject) {
         return value.isObject() ? PerfectMatchScore : IncompatibleScore;
-    } else if (QMetaType::typeFlags(targetType) & QMetaType::PointerToQObject) {
+    } else if (target.flags() & QMetaType::PointerToQObject) {
         if (value.isNull())
             return PerfectMatchScore;
         if (!value.isObject())
@@ -661,7 +799,7 @@ int QMetaObjectPublisher::conversionScore(const QJsonValue &value, int targetTyp
             return IncompatibleScore;
 
         QObject *unwrappedObject = unwrapObject(object[KEY_ID].toString());
-        return unwrappedObject != Q_NULLPTR ? PerfectMatchScore : IncompatibleScore;
+        return unwrappedObject != nullptr ? PerfectMatchScore : IncompatibleScore;
     } else if (targetType == QMetaType::QVariant) {
         return VariantScore;
     }
@@ -677,7 +815,7 @@ int QMetaObjectPublisher::conversionScore(const QJsonValue &value, int targetTyp
     QVariant variant = value.toVariant();
     if (variant.userType() == targetType) {
         return PerfectMatchScore;
-    } else if (variant.canConvert(targetType)) {
+    } else if (variant.canConvert(target)) {
         return GenericConversionScore;
     }
 
@@ -698,10 +836,10 @@ void QMetaObjectPublisher::transportRemoved(QWebChannelAbstractTransport *transp
     auto it = transportedWrappedObjects.find(transport);
     // It is not allowed to modify a container while iterating over it. So save
     // objects which should be removed and call objectDestroyed() on them later.
-    QVector<QObject*> objectsForDeletion;
+    QList<QObject *> objectsForDeletion;
     while (it != transportedWrappedObjects.end() && it.key() == transport) {
         if (wrappedObjects.contains(it.value())) {
-            QVector<QWebChannelAbstractTransport*> &transports = wrappedObjects[it.value()].transports;
+            QList<QWebChannelAbstractTransport *> &transports = wrappedObjects[it.value()].transports;
             transports.removeOne(transport);
             if (transports.isEmpty())
                 objectsForDeletion.append(wrappedObjects[it.value()].object);
@@ -712,7 +850,7 @@ void QMetaObjectPublisher::transportRemoved(QWebChannelAbstractTransport *transp
 
     transportedWrappedObjects.remove(transport);
 
-    foreach (QObject *obj, objectsForDeletion)
+    for (QObject *obj : std::as_const(objectsForDeletion))
         objectDestroyed(obj);
 }
 
@@ -747,21 +885,28 @@ QJsonValue QMetaObjectPublisher::wrapResult(const QVariant &result, QWebChannelA
                 if (oi.transports.isEmpty())
                     oi.transports = webChannel->d_func()->transports;
 
-                for (auto transport : qAsConst(oi.transports)) {
+                for (auto transport : std::as_const(oi.transports)) {
                     transportedWrappedObjects.insert(transport, id);
                 }
             }
             wrappedObjects.insert(id, oi);
 
             initializePropertyUpdates(object, classInfo);
-        } else if (wrappedObjects.contains(id)) {
-            Q_ASSERT(object == wrappedObjects.value(id).object);
-            // check if this transport is already assigned to the object
-            if (transport && !wrappedObjects.value(id).transports.contains(transport)) {
-                wrappedObjects[id].transports.append(transport);
-                transportedWrappedObjects.insert(transport, id);
+        } else {
+            auto oi = wrappedObjects.find(id);
+            if (oi != wrappedObjects.end() && !oi->isBeingWrapped) {
+                Q_ASSERT(object == oi->object);
+                // check if this transport is already assigned to the object
+                if (transport && !oi->transports.contains(transport)) {
+                    oi->transports.append(transport);
+                    transportedWrappedObjects.insert(transport, id);
+                }
+                // QTBUG-84007: Block infinite recursion for self-contained objects
+                // which have already been wrapped
+                oi->isBeingWrapped = true;
+                classInfo = classInfoForObject(object, transport);
+                oi->isBeingWrapped = false;
             }
-            classInfo = classInfoForObject(object, transport);
         }
 
         QJsonObject objectInfo;
@@ -771,7 +916,7 @@ QJsonValue QMetaObjectPublisher::wrapResult(const QVariant &result, QWebChannelA
             objectInfo[KEY_DATA] = classInfo;
 
         return objectInfo;
-    } else if (QMetaType::typeFlags(result.userType()).testFlag(QMetaType::IsEnumeration)) {
+    } else if (result.metaType().flags().testFlag(QMetaType::IsEnumeration)) {
         return result.toInt();
     } else if (isQFlagsType(result.userType())) {
         return *reinterpret_cast<const int*>(result.constData());
@@ -783,15 +928,27 @@ QJsonValue QMetaObjectPublisher::wrapResult(const QVariant &result, QWebChannelA
         // TODO: Improve QJSValue-QJsonValue conversion in Qt.
         return wrapResult(result.value<QJSValue>().toVariant(), transport, parentObjectId);
 #endif
+    } else if (result.metaType().id() == QMetaType::QString ||
+               result.metaType().id() == QMetaType::QByteArray) {
+        // avoid conversion to QVariantList
     } else if (result.canConvert<QVariantList>()) {
         // recurse and potentially wrap contents of the array
         // *don't* use result.toList() as that *only* works for QVariantList and QStringList!
         // Also, don't use QSequentialIterable (yet), since that seems to trigger QTBUG-42016
         // in certain cases.
-        return wrapList(result.value<QVariantList>(), transport);
+        // additionally, when there's a direct converter to QVariantList, use that one via convert
+        // but recover when conversion fails and fall back to the .value<QVariantList> conversion
+        // see also: https://bugreports.qt.io/browse/QTBUG-80751
+        auto list = result;
+        if (!list.convert(QMetaType::fromType<QVariantList>()))
+            list = result;
+        return wrapList(list.value<QVariantList>(), transport);
     } else if (result.canConvert<QVariantMap>()) {
         // recurse and potentially wrap contents of the map
-        return wrapMap(result.toMap(), transport);
+        auto map = result;
+        if (!map.convert(QMetaType::fromType<QVariantMap>()))
+            map = result;
+        return wrapMap(map.value<QVariantMap>(), transport);
     }
 
     return QJsonValue::fromVariant(result);
@@ -800,7 +957,7 @@ QJsonValue QMetaObjectPublisher::wrapResult(const QVariant &result, QWebChannelA
 QJsonArray QMetaObjectPublisher::wrapList(const QVariantList &list, QWebChannelAbstractTransport *transport, const QString &parentObjectId)
 {
     QJsonArray array;
-    foreach (const QVariant &arg, list) {
+    for (const QVariant &arg : list) {
         array.append(wrapResult(arg, transport, parentObjectId));
     }
     return array;
@@ -831,11 +988,57 @@ void QMetaObjectPublisher::broadcastMessage(const QJsonObject &message) const
         return;
     }
 
-    foreach (QWebChannelAbstractTransport *transport, webChannel->d_func()->transports) {
+    for (QWebChannelAbstractTransport *transport : webChannel->d_func()->transports) {
         transport->sendMessage(message);
     }
 }
 
+void QMetaObjectPublisher::enqueueBroadcastMessage(const QJsonObject &message)
+{
+    if (webChannel->d_func()->transports.isEmpty()) {
+        qWarning("QWebChannel is not connected to any transports, cannot send message: %s",
+                 QJsonDocument(message).toJson().constData());
+        return;
+    }
+
+    for (auto *transport : webChannel->d_func()->transports) {
+        auto &state = transportState[transport];
+        state.queuedMessages.append(message);
+    }
+}
+
+void QMetaObjectPublisher::enqueueMessage(const QJsonObject &message,
+                                          QWebChannelAbstractTransport *transport)
+{
+    auto &state = transportState[transport];
+    state.queuedMessages.append(message);
+}
+
+void QMetaObjectPublisher::sendEnqueuedPropertyUpdates(QWebChannelAbstractTransport *transport)
+{
+    auto found = transportState.find(transport);
+    if (found != transportState.end() && found.value().clientIsIdle
+        && !found.value().queuedMessages.isEmpty()) {
+
+        // If the client is connected with an in-process transport, it can
+        // happen that a message triggers a subsequent property change. In
+        // that case, we need to ensure that the queued messages have already
+        // been cleared; otherwise the recursive call will send everythig again.
+        // Case in point: The qmlwebchannel tests fail if we don't clear the
+        // queued messages before sending them out.
+        // For that same reason set the client to "busy" (aka non-idle) just
+        // right before sending out the messages; otherwise a potential
+        // "Idle" type message will not correctly restore the Idle state.
+        const auto messages = std::move(found.value().queuedMessages);
+        Q_ASSERT(found.value().queuedMessages.isEmpty());
+        found.value().clientIsIdle = false;
+
+        for (const auto &message : messages) {
+            transport->sendMessage(message);
+        }
+    }
+}
+
 void QMetaObjectPublisher::handleMessage(const QJsonObject &message, QWebChannelAbstractTransport *transport)
 {
     if (!webChannel->d_func()->transports.contains(transport)) {
@@ -850,7 +1053,7 @@ void QMetaObjectPublisher::handleMessage(const QJsonObject &message, QWebChannel
 
     const MessageType type = toType(message.value(KEY_TYPE));
     if (type == TypeIdle) {
-        setClientIsIdle(true);
+        setClientIsIdle(true, transport);
     } else if (type == TypeInit) {
         if (!message.contains(KEY_ID)) {
             qWarning("JSON message object is missing the id property: %s",
@@ -860,7 +1063,7 @@ void QMetaObjectPublisher::handleMessage(const QJsonObject &message, QWebChannel
         transport->sendMessage(createResponse(message.value(KEY_ID), initializeClient(transport)));
     } else if (type == TypeDebug) {
         static QTextStream out(stdout);
-        out << "DEBUG: " << message.value(KEY_DATA).toString() << endl;
+        out << "DEBUG: " << message.value(KEY_DATA).toString() << Qt::endl;
     } else if (message.contains(KEY_OBJECT)) {
         const QString &objectName = message.value(KEY_OBJECT).toString();
         QObject *object = registeredObjects.value(objectName);
@@ -893,13 +1096,33 @@ void QMetaObjectPublisher::handleMessage(const QJsonObject &message, QWebChannel
                                       method.toInt(-1),
                                       message.value(KEY_ARGS).toArray());
             }
-            if (!publisherExists || !transportExists)
-                return;
-            transport->sendMessage(createResponse(message.value(KEY_ID), wrapResult(result, transport)));
+
+            auto sendResponse = [publisherExists, transportExists, id=message.value(KEY_ID)]
+                    (const QVariant &result)
+            {
+                if (!publisherExists || !transportExists)
+                    return;
+
+                Q_ASSERT(QThread::currentThread() == publisherExists->thread());
+
+                const auto wrappedResult =
+                        publisherExists->wrapResult(result, transportExists.get());
+                transportExists->sendMessage(createResponse(id, wrappedResult));
+            };
+
+#if QT_CONFIG(future)
+            if (result.canConvert<QFuture<void>>()) {
+                attachContinuationToFutureInVariant(result, publisherExists.get(), sendResponse);
+            } else {
+                sendResponse(result);
+            }
+#else
+            sendResponse(result);
+#endif
         } else if (type == TypeConnectToSignal) {
-            signalHandler.connectTo(object, message.value(KEY_SIGNAL).toInt(-1));
+            signalHandlerFor(object)->connectTo(object, message.value(KEY_SIGNAL).toInt(-1));
         } else if (type == TypeDisconnectFromSignal) {
-            signalHandler.disconnectFrom(object, message.value(KEY_SIGNAL).toInt(-1));
+            signalHandlerFor(object)->disconnectFrom(object, message.value(KEY_SIGNAL).toInt(-1));
         } else if (type == TypeSetProperty) {
             setProperty(object, message.value(KEY_PROPERTY).toInt(-1),
                         message.value(KEY_VALUE));
@@ -907,29 +1130,58 @@ void QMetaObjectPublisher::handleMessage(const QJsonObject &message, QWebChannel
     }
 }
 
+int QMetaObjectPublisher::propertyUpdateInterval()
+{
+    return propertyUpdateIntervalTime;
+}
+
+void QMetaObjectPublisher::setPropertyUpdateInterval(int ms)
+{
+    propertyUpdateIntervalTime = ms;
+}
+
 void QMetaObjectPublisher::setBlockUpdates(bool block)
 {
-    if (blockUpdates == block) {
-        return;
-    }
-    blockUpdates = block;
+    blockUpdatesStatus = block;
+}
 
-    if (!blockUpdates) {
+void QMetaObjectPublisher::onBlockUpdatesChanged()
+{
+    if (!blockUpdatesStatus) {
+        startPropertyUpdateTimer();
         sendPendingPropertyUpdates();
     } else if (timer.isActive()) {
         timer.stop();
     }
 
-    emit blockUpdatesChanged(block);
+    emit blockUpdatesChanged(blockUpdatesStatus);
+}
+
+bool QMetaObjectPublisher::blockUpdates() const
+{
+    return blockUpdatesStatus;
 }
 
 void QMetaObjectPublisher::timerEvent(QTimerEvent *event)
 {
     if (event->timerId() == timer.timerId()) {
+        if (propertyUpdateIntervalTime <= 0)
+            timer.stop();
         sendPendingPropertyUpdates();
     } else {
         QObject::timerEvent(event);
     }
 }
 
+SignalHandler<QMetaObjectPublisher> *QMetaObjectPublisher::signalHandlerFor(const QObject *object)
+{
+    auto thread = object->thread();
+    auto it = signalHandlers.find(thread);
+    if (it == signalHandlers.end()) {
+        it = signalHandlers.emplace(thread, this).first;
+        it->second.moveToThread(thread);
+    }
+    return &it->second;
+}
+
 QT_END_NAMESPACE
